@@ -291,6 +291,7 @@ public:
   double unsplit_oper_cost;
   /* Cardinality of T when nothing is pushed */
   double unsplit_card;
+  double last_refills;
 
   SplM_plan_info *find_plan(TABLE *table, uint key, uint parts);
 };
@@ -509,9 +510,7 @@ bool JOIN::check_for_splittable_materialized()
     the collected info on potential splittability of T
   */
   SplM_opt_info *spl_opt_info= new (thd->mem_root) SplM_opt_info();
-  SplM_field_info *spl_field=
-    (SplM_field_info *) (thd->calloc(sizeof(SplM_field_info) *
-                                            spl_field_cnt));
+  SplM_field_info *spl_field= thd->calloc<SplM_field_info>(spl_field_cnt);
 
   if (!(spl_opt_info && spl_field)) // consider T as not good for splitting
     return false;
@@ -604,29 +603,17 @@ void TABLE::add_splitting_info_for_key_field(KEY_FIELD *key_field)
   THD *thd= in_use;
   Item *left_item= spl_field->producing_item->build_clone(thd);
   Item *right_item= key_field->val->build_clone(thd);
-  Item_bool_func *eq_item= 0;
+  Item_func_eq *eq_item= 0;
   if (left_item && right_item)
   {
     right_item->walk(&Item::set_fields_as_dependent_processor,
                      false, join->select_lex);
     right_item->update_used_tables();
-    /*
-      We've just pushed right_item down into the child select. It may only
-      have references to outside.
-    */
-    DBUG_ASSERT(!(right_item->used_tables() & ~PSEUDO_TABLE_BITS));
-
-    //  Item_func::EQUAL_FUNC is null-safe, others can use Item_func_eq()
-    if (key_field->cond->type() == Item::FUNC_ITEM &&
-        ((Item_func*)key_field->cond)->functype() == Item_func::EQUAL_FUNC)
-      eq_item= new (thd->mem_root) Item_func_equal(thd, left_item, right_item);
-    else
-      eq_item= new (thd->mem_root) Item_func_eq(thd, left_item, right_item);
+    eq_item= new (thd->mem_root) Item_func_eq(thd, left_item, right_item);
   }
   if (!eq_item)
     return;
-  KEY_FIELD *added_key_field=
-    (KEY_FIELD *) thd->alloc(sizeof(KEY_FIELD));
+  KEY_FIELD *added_key_field= thd->alloc<KEY_FIELD>(1);
   if (!added_key_field ||
       spl_opt_info->added_key_fields.push_back(added_key_field,thd->mem_root))
     return;
@@ -636,7 +623,14 @@ void TABLE::add_splitting_info_for_key_field(KEY_FIELD *key_field)
   added_key_field->level= 0;
   added_key_field->optimize= KEY_OPTIMIZE_EQ;
   added_key_field->eq_func= true;
-  added_key_field->null_rejecting= key_field->null_rejecting;
+
+  Item *real= key_field->val->real_item();
+  if ((real->type() == Item::FIELD_ITEM) &&
+        ((Item_field*)real)->field->maybe_null())
+    added_key_field->null_rejecting= true;
+  else
+    added_key_field->null_rejecting= false;
+
   added_key_field->cond_guard= NULL;
   added_key_field->sj_pred_no= UINT_MAX;
   return;
@@ -1060,25 +1054,6 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
   if (best_table)
   {
     *spl_pd_boundary= this->table->map;
-    /*
-      Compute "refills" - how many times we'll need to refill the split-
-      materialized temp. table. Split-materialized table has references to
-      preceding table(s). Suppose the join prefix is (t1, t2, t3) and
-      split-materialized refers to table t2:
-
-        t1  t2  t3  <split_materialized>
-            ^             |
-            +------------ +
-
-      If we do not use join buffer for table t3, then we'll need to refill
-      the split-materialized table partial_join_cardinality({t1, t2}) times.
-      (this assumes that fanout of table t3 is greater than 1, which is
-      typically true).
-      If table t3 uses join buffer, then every time we get a record combination
-      of {t1.row,t2.row,t3.row} the t2.row may be different and so we will need
-      to refill <split_materialized> every time, that is,
-      partial_join_cardinality(t1,t3,t3) times.
-    */
     if (!best_param_tables)
       refills= 1;
     else
@@ -1130,9 +1105,8 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
       key_map spl_keys= table->keys_usable_for_splitting;
       if (!(first_non_const_pos->key &&
             spl_keys.is_set(first_non_const_pos->key->key)) ||
-          !(spl_plan= (SplM_plan_info *) thd->alloc(sizeof(SplM_plan_info))) ||
-	  !(spl_plan->best_positions=
-	     (POSITION *) thd->alloc(sizeof(POSITION) * join->table_count)) ||
+          !(spl_plan= thd->alloc<SplM_plan_info>(1)) ||
+	  !(spl_plan->best_positions= thd->alloc<POSITION>(join->table_count)) ||
 	  spl_opt_info->plan_cache.push_back(spl_plan))
       {
         reset_validity_vars_for_keyuses(best_key_keyuse_ext_start, best_table,
@@ -1199,7 +1173,7 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
       The best plan that employs splitting is cheaper than
       the plan without splitting
     */
-    startup_cost= refills * spl_plan->cost;
+    startup_cost= spl_opt_info->last_refills * spl_plan->cost;
     records= (ha_rows) (spl_opt_info->unsplit_card * spl_plan->split_sel);
     if (unlikely(thd->trace_started()) && ! already_printed)
     {
